/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2010, James Martelletti <james@nerdc0re.com>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * James Martelletti <james@nerdc0re.com>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * James Martelletti <james@nerdc0re.com>
 *
 *
 * mod_journal.c -- System Logging
 *
 */
#include <switch.h>
#include <stdlib.h>
#include <systemd/sd-journal.h>
#include <sys/uio.h>

#define DEFAULT_IDENT    "freeswitch"
#define DEFAULT_FACILITY LOG_USER
#define DEFAULT_LEVEL    "warning"
#define DEFAULT_FORMAT   "[message]"
#define MAX_LENGTH       1024

SWITCH_MODULE_LOAD_FUNCTION(mod_journal_load);
SWITCH_MODULE_SHUTDOWN_FUNCTION(mod_journal_shutdown);
SWITCH_MODULE_DEFINITION(mod_journal, mod_journal_load, mod_journal_shutdown, NULL);

static switch_status_t load_config(void);


static struct {
	char *ident;
	char *format;
	int facility;
	switch_bool_t log_uuid;
	switch_event_t *session_vars;
	switch_log_level_t log_level;
} globals;

struct _facility_table_entry {
	char *description;
	int facility;
};
static switch_log_json_format_t json_format1 = {
	{ NULL, NULL }, // version
	{ "host", NULL }, // host
	{ "timestamp", NULL }, // timestamp
	{ "level", NULL }, // level
	{ "ident", NULL }, // ident
	{ "pid", NULL }, // pid
	{ "session", NULL }, // uuid
	{ "file", NULL }, // file
	{ "line", NULL }, // line
	{ "function", NULL }, // function
	{ "message", NULL }, // full_message
	{ NULL, NULL }, // short_message
	"", // custom_field_prefix
	1000000.0, // timestamp_divisor
	{ "sequence", NULL } // sequence
};
static switch_log_json_format_t json_format = {
	{ NULL, NULL }, // version
	{ NULL, NULL }, // host
	{ NULL, NULL }, // timestamp
	{ NULL, NULL }, // level
	{ NULL, NULL }, // ident
	{ NULL, NULL }, // pid
	{ NULL, NULL }, // uuid
	{ NULL, NULL }, // file
	{ NULL, NULL }, // line
	{ NULL, NULL }, // function
	{ "message", NULL }, // full_message
	{ "message2", NULL }, // short_message
	"", // custom_field_prefix
	1000000.0, // timestamp_divisor
	{ NULL, NULL } // sequence
};
SWITCH_DECLARE_GLOBAL_STRING_FUNC(set_global_ident, globals.ident);
SWITCH_DECLARE_GLOBAL_STRING_FUNC(set_global_format, globals.format);

switch_status_t set_global_facility(const char *facility)
{
	const struct _facility_table_entry facilities[] = {
		{"auth", LOG_AUTH},
#if !defined (__SVR4) && !defined (__sun)
		{"authpriv", LOG_AUTHPRIV},
		{"ftp", LOG_FTP},
#endif
		{"cron", LOG_CRON},
		{"daemon", LOG_DAEMON},
		{"kern", LOG_KERN},
		{"local0", LOG_LOCAL0},
		{"local1", LOG_LOCAL1},
		{"local2", LOG_LOCAL2},
		{"local3", LOG_LOCAL3},
		{"local4", LOG_LOCAL4},
		{"local5", LOG_LOCAL5},
		{"local6", LOG_LOCAL6},
		{"local7", LOG_LOCAL7},
		{"lpr", LOG_LPR},
		{"mail", LOG_MAIL},
		{"news", LOG_NEWS},
		{"syslog", LOG_SYSLOG},
		{"user", LOG_USER},
		{"uucp", LOG_UUCP},
		{NULL, 0}
	};
	const struct _facility_table_entry *entry = facilities;

	while (!zstr(entry->description)) {
		if (!strcasecmp(entry->description, facility)) {
			globals.facility = entry->facility;
			return SWITCH_STATUS_SUCCESS;
		}
		entry++;
	}

	return SWITCH_STATUS_FALSE;
}

static switch_loadable_module_interface_t console_module_interface = {
	/*.module_name */ modname,
	/*.endpoint_interface */ NULL,
	/*.timer_interface */ NULL,
	/*.dialplan_interface */ NULL,
	/*.codec_interface */ NULL,
	/*.application_interface */ NULL,
	/*.api_interface */ NULL,
	/*.file_interface */ NULL,
	/*.speech_interface */ NULL,
	/*.directory_interface */ NULL
};


static int find_unprintable(const char *s)
{
	const char *p;

	for(p = s; p && *p; p++) {
		if (*p < 9 || *p == 27) {
			return 1;
		}
	}

	return 0;
}

static switch_status_t mod_journal_logger(const switch_log_node_t *node, switch_log_level_t level)
{
	const int N_MAX = 50;
	int syslog_level;
	char line_buf[24];
	char file_buf[100];
	switch_core_session_t *session = NULL;
	switch_event_header_t *hp;
	char *data_tmp = NULL;
	struct iovec iov[N_MAX];
	int n = 0;

	switch (level) {
	case SWITCH_LOG_DEBUG:
		syslog_level = LOG_DEBUG;
		break;
	case SWITCH_LOG_INFO:
		syslog_level = LOG_INFO;
		break;
	case SWITCH_LOG_NOTICE:
		syslog_level = LOG_NOTICE;
		break;
	case SWITCH_LOG_WARNING:
		syslog_level = LOG_WARNING;
		break;
	case SWITCH_LOG_ERROR:
		syslog_level = LOG_ERR;
		break;
	case SWITCH_LOG_CRIT:
		syslog_level = LOG_CRIT;
		break;
	case SWITCH_LOG_ALERT:
		syslog_level = LOG_ALERT;
		break;
	default:
		syslog_level = LOG_NOTICE;
		break;
	}
	// typedef struct {
	//     /*! The complete log message */
	//     char *data;
	//     /*! The file where the message originated */
	//     char file[80];
	//     /*! The line number where the message originated */
	//     uint32_t line;
	//     /*! The function where the message originated */
	//     char func[80];
	//     /*! The log level of the message */
	//     switch_log_level_t level;
	//     /*! The time when the log line was sent */
	//     switch_time_t timestamp;
	//     /*! A pointer to where the actual content of the message starts (skipping past the preformatted portion) */
	//     char *content;
	//     char *userdata;
	//     /* To maintain abi, only add new elements to the end of this struct and do not delete any elements */
	//     switch_text_channel_t channel;
	//     switch_log_level_t slevel;
	//     switch_event_t *tags;
	//     /* Log sequence */
	//     int64_t sequence;
	//     /* Optional extra log metadata */
	//     cJSON *meta;
	// } switch_log_node_t;
	snprintf(line_buf, sizeof(line_buf), "CODE_LINE=%d", node->line);
	snprintf(file_buf, sizeof(file_buf), "CODE_FILE=%s", node->file);

	//The content sometimes starts with a space, so ignore it (It seems like it doesn't start with a space if the clean log is used instead of the normal one.)
	iov[n].iov_len = asprintf(&data_tmp, "MESSAGE=%s", node->content + (*node->content == ' '));
	iov[n].iov_base = data_tmp;
	n++;
	iov[n].iov_len = asprintf(&data_tmp, "PRIORITY=%i", syslog_level);
	iov[n].iov_base = data_tmp;
	n++;
	iov[n].iov_len = asprintf(&data_tmp, "SYSLOG_RAW=%s", node->data);
	iov[n].iov_base = data_tmp;
	n++;
	iov[n].iov_len = asprintf(&data_tmp, "TIMESTAMP=%llu", node->timestamp);
	iov[n].iov_base = data_tmp;
	n++;
	iov[n].iov_len = asprintf(&data_tmp, "HAS_SESSION=%s", node->userdata?"true":"false");
	iov[n].iov_base = data_tmp;
	n++;
	iov[n].iov_len = asprintf(&data_tmp, "LOG_CHANNEL=%i", node->channel);
	iov[n].iov_base = data_tmp;
	n++;
	iov[n].iov_len = asprintf(&data_tmp, "HAS_TAGS=%s", node->tags?"true":"false");
	iov[n].iov_base = data_tmp;
	n++;

	if(!zstr(node->userdata)){
		iov[n].iov_len = asprintf(&data_tmp, "SESSION_UUID=%i", node->userdata);
		iov[n].iov_base = data_tmp;
		n++;
	}
	if (!node->tags && (session = switch_core_session_locate(node->userdata))) {
		const char *domain = NULL;
		const char *user = NULL;

		switch_channel_t *channel = switch_core_session_get_channel(session);
		domain = switch_channel_get_variable_dup(channel, "domain_name", SWITCH_FALSE, -1);
		if(!zstr(domain)){
			iov[n].iov_len = asprintf(&data_tmp, "DOMAIN_NAME=%s", domain);
			iov[n].iov_base = data_tmp;
			n++;
		}
		user = switch_channel_get_variable_dup(channel, "user_name", SWITCH_FALSE, -1);
		if(!zstr(user)){
			iov[n].iov_len = asprintf(&data_tmp, "USER_NAME=%s", user);
			iov[n].iov_base = data_tmp;
			n++;
		}
		switch_core_session_rwunlock(session);
	}
	if(node->tags){
		for (hp = node->tags->headers; hp; hp = hp->next) {
			if(n>=N_MAX){break;}
			if (!zstr(hp->name) && !zstr(hp->value)) {
				char* name = switch_uc_strdup(hp->name);
				iov[n].iov_len = asprintf(&data_tmp, "%s=%s", name, hp->value);
				iov[n].iov_base = data_tmp;
				n++;
				free(name);
			}
		}
	}

	sd_journal_sendv_with_location(file_buf, line_buf, node->func, iov, n);
	while(n--){
		switch_safe_free(iov[n].iov_base);
	}

	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t load_config(void)
{
	char *cf = "journal.conf";
	switch_xml_t cfg, xml, settings, param;

	/* default log level */
	globals.log_level = SWITCH_LOG_DEBUG;

	/* default facility */
	globals.facility = DEFAULT_FACILITY;
	globals.log_uuid = SWITCH_TRUE;

	if (!(xml = switch_xml_open_cfg(cf, &cfg, NULL))) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Open of %s failed\n", cf);
	} else {
		if ((settings = switch_xml_child(cfg, "settings"))) {
			for (param = switch_xml_child(settings, "param"); param; param = param->next) {
				char *var = (char *) switch_xml_attr_soft(param, "name");
				char *val = (char *) switch_xml_attr_soft(param, "value");

				if (!strcmp(var, "ident")) {
					set_global_ident(val);
				} else if (!strcmp(var, "format")) {
					set_global_format(val);
				} else if (!strcmp(var, "facility")) {
					set_global_facility(val);
				} else if (!strcasecmp(var, "loglevel") && !zstr(val)) {
					globals.log_level = switch_log_str2level(val);
					if (globals.log_level == SWITCH_LOG_INVALID) {
						globals.log_level = SWITCH_LOG_WARNING;
					}
				} else if (!strcasecmp(var, "uuid")) {
					globals.log_uuid = switch_true(val);
				} else if (!strcasecmp(var, "session-vars")) {
					switch_event_create_json(&globals.session_vars, val);
				}
			}
		}
		switch_xml_free(xml);
	}

	if (zstr(globals.ident)) {
		set_global_ident(DEFAULT_IDENT);
	}
	if (zstr(globals.format)) {
		set_global_format(DEFAULT_FORMAT);
	}
	return 0;
}

SWITCH_MODULE_LOAD_FUNCTION(mod_journal_load)
{
	switch_status_t status;
	*module_interface = &console_module_interface;

	memset(&globals, 0, sizeof(globals));

	if ((status = load_config()) != SWITCH_STATUS_SUCCESS) {
		return status;
	}
	//TODO: load the is_console from the config file or something
	switch_log_bind_logger(mod_journal_logger, globals.log_level, SWITCH_FALSE);

	return SWITCH_STATUS_SUCCESS;
}

SWITCH_MODULE_SHUTDOWN_FUNCTION(mod_journal_shutdown)
{
	switch_safe_free(globals.ident);
	switch_safe_free(globals.format);
	if(globals.session_vars) switch_event_destroy(&globals.session_vars);

	switch_log_unbind_logger(mod_journal_logger);

	return SWITCH_STATUS_SUCCESS;
}

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 noet:
 */
